import numpy as np
from scipy.special import factorial, factorial2
from scipy.special import gamma as sciGamma

from thecobs.Constants import *

def gamma_k(k, Z):
    return (k*k-(ALPHA*Z)**2)**0.5

def generalizedFermiFunction(W, Z, R, k):
    """Implementation of the generalized Fermi function F_{k-1} according to Behrens et al.

    :param W: Total electron energy in units of its rest mass
    :param Z: Proton number of daughter
    :param R: nuclear radius in natural units
    :param k: absolute value of kappa

    """
    p = np.sqrt(W**2.0-1.)
    y = ALPHA*Z*W/p
    gk = gamma_k(k, Z)
    prefactor = (k*factorial2(2*k-1))**2.0*4**k*(2*p*R)**(2.*(gk-k))*np.exp(np.pi*y)
    gammas = (np.absolute(sciGamma(gk+y*1j))/sciGamma(1+2.*gk))**2.0
    return prefactor*gammas

def lambda_2(W, Z, R):
    p = np.sqrt(W**2.0-1.)
    y = ALPHA*Z*W/p
    g2 = gamma_k(2, Z)
    g1 = gamma_k(1, Z)
    prefactor = (2+g2)/2/(1+g1)/(1+(ALPHA*Z)**2.0/12*(11.577216))
    kinematic = (2*p*R)**(2*(g2-g1-1))*(np.absolute(sciGamma(g2+y*1j))/np.absolute(sciGamma(g1+y*1j)))**2.0
    return prefactor*kinematic

def lambda_k(W, Z, R, k):
    """Coulomb function $\lambda_k$ as per Behrens et al.

    :param W: Total electron energy in units of its rest mass
    :param Z: Proton number of daughter
    :param k: absolute value of kappa

    """
    gk = gamma_k(k, Z)
    g1 = gamma_k(1, Z)
    return generalizedFermiFunction(W, Z, R, k)/generalizedFermiFunction(W, Z, R, 1)*(k+gk)/(k*(1+g1))

def calcQ_k(k, Z, W, beta):
    """Calculate the Q screening ratio according to Buehring (1984)

    :param k: absolute value of kappa
    :param Z: Proton number of daughter
    :param W: Total electron energy in units of its rest mass
    :param beta: screening exponent at r=0

    """
    gk = gamma_k(k, Z)

    up = (ALPHA * Z) ** 2.0 * calcN2(-k, Z, W, beta) + (k + gk) ** 2.0 * calcN2(k, Z, W, beta)

    down = (ALPHA * Z) ** 2.0 * calcNCoul2(-k, Z, W) + (k + gk) ** 2.0 * calcNCoul2(k, Z, W)

    return up / down


def calcN2(kappa, Z, W, beta):
    """Calculate the square of the normalization function according to Buehring (1984)

    :param k: absolute value of kappa
    :param Z: Proton number of daughter
    :param W: Total electron energy in units of its rest mass
    :param beta: screening exponent at r=0
    :param kappa:

    """
    k = abs(kappa)

    Wt = W - ALPHA * Z * beta / 2.0

    p = sqrt(W ** 2.0 - 1.0)

    pt = 0.5 * p + 0.5 * sqrt(p ** 2.0 - 2 * ALPHA * Z * beta * Wt + 0.0 * (1j))

    kappat = kappa * sqrt(1.0 - ALPHA * Z * beta / (1.0 + W) + 0.0 * 1j)

    P = p / beta
    Pt = pt / beta

    gk = gamma_k(k, Z)

    yt = ALPHA * Z * Wt / pt

    term1 = 0.5 * ((kappa - gk) * (kappa * Wt - gk) + 0.5 * (gk / kappa) * (kappat - kappa) ** 2.0)
    R_ab = 1.0
    if (kappa > 0):
        R_ab += 0.25 * k ** 2.0 * (beta / p) ** 2.0

    c1 = 1.0 + 2.0 * gamma_k
    c2 = gk + yt * (1j)
    c3 = gk + 2.0 * Pt * (1j)
    c4 = k + 2.0 * P * (1j)

    Gc1 = fabs(gamma(c1))
    Gc2 = fabs(gamma(c2))

    Gc3 = log(gamma(c3))
    Gc4 = log(gamma(c4))

    Gc34 = exp(Gc3 - Gc4)

    rap2 = (abs(Gc34)) ** 2.0

    if c3.imag > 200.0 and c4.imag > 200.0:
        crap2 = sqrt(c4/c3)*exp(c4-c3)
        lnrap = c3*log(c3)-c4*log(c4)
        rap2 = (abs(crap2*exp(lnrap)))**2.0

    rap1 = (Gc2 / Gc1) ** 2.0

    rap = rap1 * rap2

    term5 = beta ** (2.0 * gk - 2.0 * k) * (2.0 * p) ** (2.0 * k - 1.0)

    return term1 * (rap / R_ab) * term5 + 0.0 * 1j


def calcNCoul2(kappa, Z, W):
    """Calculate the square of the normalization function for a simple point charge, according to Buehring (1984)

    :param kappa: absolute value of kappa
    :param Z: Proton number of daughter
    :param W: Total electron energy in units of its rest mass

    """
    k = abs(kappa)

    gk = gamma_k(k, Z)

    p = sqrt(W ** 2.0 - 1.0)

    y = ALPHA * Z * W / p

    return 0.5 * (kappa - gk) * (kappa * W - gk) * (gamma(1.0 + 2.0 * gk)) ** (-2.0) * (
        np.absolute(gamma(gk + y * 1j))) ** 2.0 * exp(pi * y) * (2.0 * p) ** (
                       2.0 * gk - 1.0) + 0.0 * 1j

def calcLambda_k_scr(k, Z, W, beta):
    """Calculate the screening correction to the $\lambda_k$ Coulomb function

    :param k: absolute value of kappa
    :param Z: Proton number of daughter
    :param W: Total electron energy in units of its rest mass
    :param beta: Screening exponent at r=0

    """
    return calcQ_k(k, Z, W, beta)/calcQ_k(1.0, Z, W, beta)



import numpy as np
from scipy.interpolate import interp1d

from thecobs.Constants import *

def stableA(Z):
    '''Get estimate for stable mass number as a function of Z'''

    '''Parameters from Bethe-Weiszacker formula'''
    aC = 0.71 # MeV
    aA = 23.7 # MeV
    alpha = aC/4/aA

    return 2*Z*(1+alpha*(2*Z)**(2/3))

def getEltonNuclearRadius(A, natUnits=False):
    """Return nuclear charge radius in fm according to Elton formula

    :param A: mass number
    :param natUnits: return result in units where m_e=hbar=c=1 (Default value = False)

    """
    R = 1.121 * A ** (1. / 3.) + 2.426 * A ** (-1. / 3.) - 6.614 / A

    if natUnits:
        """ hbar * c / m_e, in fm"""
        R /= NATURAL_LENGTH/1e-15
    return R

def getDJpi(Jpi):
    """Turn string of possible spin-parities into
    (double of spin, parity)

    :param Jpi: string, spin-parities

    """
    parity = 1
    DJ = 0
    if '-' in Jpi:
        parity = -1
        DJ = Jpi.split('-')[0]
    else:
        DJ = Jpi.split('+')[0]
    try:
        if '/2' in Jpi:
            DJ = int(DJ.split('/2')[0])
        else:
            DJ = 2 * int(DJ)
    except ValueError:
        print(Jpi)
        print(parity)
        print(DJ)
        raise
        return (np.nan, np.nan)
    return (DJ, parity)


def determineForbiddenness(mJpi, dJpi):
    """Determine degree of forbiddenness and uniqueness
    returns (deltaJ[int], forbiddenness[int], unique[bool])

    :param mJpi: string, mother spin and parity
    :param dJpi: string, daughter spin and parity

    """
    mDJpi = getDJpi(mJpi)
    dDJpi = getDJpi(dJpi)

    unique = False

    deltaJ = mDJpi[1] * dDJpi[1] * abs((abs(mDJpi[0]) - abs(dDJpi[0]))) / 2
    try:
        forbiddenness = int(max(abs(deltaJ) - 1, 0))
    except ValueError:
        print(mJpi, dJpi, mDJpi, dDJpi, deltaJ)
        raise
    if np.sign(mDJpi[1] * dDJpi[1]) != (-1) ** (forbiddenness):
        forbiddenness += 1
    if abs(deltaJ) == forbiddenness + 1:
        unique = True
    return (deltaJ, forbiddenness, unique)


def rebin(x1, x2, y2):
    """Rough rebinning to downsample y2 to grid of x1

    :param x1: array, x values to rebin to
    :param x2: array, initial x values
    :param y2: array, initial y values

    """
    x1Min = np.min(x1)
    x1Max = np.max(x1)
    binSize1 = x1[1] - x1[0]
    binSize2 = x2[1] - x2[0]
    if len(y2.shape) == 1:
        rebinnedY2 = np.zeros(len(x1))
    else:
        rebinnedY2 = np.zeros((len(x1), y2.shape[1]))
    for j in range(len(y2)):
        if (x1Max + binSize1 / 2.0) > x2[j] >= (x1Min - binSize1 / 2.0):
            rebinnedY2[int((x2[j] - (x1Min + binSize1 / 2.0)) / binSize1)] += y2[j] * binSize2 / binSize1
    return rebinnedY2


def downsample(x1, x2, y2, method='linear', fillValue=0., axis=0):
    """Downsample (x2, y2) to (x1, new y)

    :param x1: array, target x range
    :param x2: array, initial x range
    :param y2: array, initial y range
    :param method: interpolation method (Default value = 'linear')
    :param fillValue: default value outside of range (Default value = 0.)

    """
    try:
        f = interp1d(x2, y2, kind=method, bounds_error=False, fill_value=fillValue, axis=axis)
        return np.array(f(x1))
    except ValueError:
        return y2



import pandas as pd
import numpy as np

from importlib import resources

from thecobs.SpectralFunctions import *
from thecobs.Constants import *

def getExchangeParams(Z):
    with resources.open_text("data", "ExchangeData.dat") as fid:
        data = np.genfromtxt(fid)
        index = np.where(data[:, 0] == Z)

        return data[index, 1:]

def calculateSpectrum(E, params):
    W = 1 + E/ELECTRON_MASS_KEV

    ph = phase_space(W, **params)
    f = fermi_function(W, **params)
    l0 = finite_size_L0(W, **params)
    u = finite_size_U_fermi(W, **params)
    rc = radiative_correction(W, **params)

    rec = np.ones(len(E))
    recCoul = np.ones(len(E))
    C = np.ones(len(E))

    beta_type = params['Type']

    if beta_type == 'A: Fermi':
        rec = recoil_fermi(W, **params)
        recCoul = recoil_Coulomb_fermi(W, **params)
        C = shape_factor_fermi(W, **params)
    elif beta_type == 'A: Gamow-Teller':
        rec = recoil_gamow_teller(W, **params)
        recCoul = recoil_Coulomb_gamow_teller(W, **params)
        C = shape_factor_gamow_teller(W, **params)
    elif beta_type == 'A: Mixed':
        norm = 1+mixing_ratio**2
        recF = recoil_fermi(W, **params)
        recCoulF = recoil_Coulomb_fermi(W, **params)
        CF = shape_factor_fermi(W, **params)
        recGT = recoil_gamow_teller(W, **params)
        recCoulGT = recoil_Coulomb_gamow_teller(W, **params)
        CGT = shape_factor_gamow_teller(W, **params)
        rec = 1+(recF-1+mixing_ratio**2*(recGT-1))/norm
        recCoul = 1+(recCoulF-1+mixing_ratio**2*(recCoulGT-1))/norm
        C = 1+(CF-1 + mixing_ratio**2*(CGT-1))/norm
    elif 'FU' in beta_type:
        C = shape_factor_unique_forbidden(W, **params)
    s = atomic_screening(W, **params)
    X = atomic_exchange(W, **params)

    sp = ph*f*l0*u*rc*rec*recCoul*C*s*X

    comb = np.stack((E, W, sp, ph, f, l0, rc, C, s, X, u, rec, recCoul), axis=1)

    df = pd.DataFrame(comb, columns = ['Energy', 'W', 'Spectrum', 'PhaseSpace', 'FermiFunction', 'L0', 'RadiativeCorrections', 'ShapeFactor', 'Screening', 'Exchange', 'U', 'Recoil', 'CoulombRecoil'])

    return df
